/*

nsisPatchGen
Copyright (C) 2006  Vibration Technology Ltd. ( http://www.vibtech.co.uk )

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*/

/**
  * \file
**/

#include "StartExternals.h"
#include <QString>
#include <QStringList>
#include <QTextStream>
#include <QDir>
#include <QFile>
#include "StopExternals.h"

#include "DirectoryDiff.h"

DirectoryDiff::DirectoryDiff( )
: mEntryListFilter( QDir::NoDotAndDotDot | QDir::NoSymLinks )
{
	QFile file( "ignore.txt" );
	if ( file.open(QIODevice::ReadOnly | QIODevice::Text) )
	{
		QTextStream in(&file);
		while ( !in.atEnd() )
			mIgnoreList.append( QRegExp( in.readLine(), Qt::CaseInsensitive, QRegExp::Wildcard ) );
	}
}

DirectoryDiff::~DirectoryDiff( )
{
}

void DirectoryDiff::compare( const QString& directory1, const QString& directory2 ) throw( QString )
{
	mDirectory1 = directory1;
	mDirectory2 = directory2;

	QString tmp = directory1;
	tmp.replace( '\\', "\\\\" );
	tmp.replace( '(', "\\(" );
	tmp.replace( ')', "\\)" );
	mBasePathRemover.setPattern( "^" + tmp + "\\\\?" );

	comparePrivate( directory1, directory2 );
}

void DirectoryDiff::comparePrivate( const QString& directory1, const QString& directory2 ) throw( QString )
{
	QDir dir1( directory1 );
	QDir dir2( directory2 );

	if( !dir1.exists() ) throw QString( directory1 + " not found." );
	if( !dir2.exists() ) throw QString( directory2 + " not found." );

	compareSubDirectoriesIn( directory1, directory2 );
	compareFilesIn( directory1, directory2 );
}



void DirectoryDiff::compareSubDirectoriesIn( const QString& directory1, const QString& directory2 )
{
	if ( isIgnored( directory1 ) )
		return;

	QDir dir1( directory1 );
	QDir dir2( directory2 );

	static QStringList emptyFileFilter;

	QStringList dirsInDir1 = dir1.entryList( emptyFileFilter, QDir::AllDirs | mEntryListFilter );
	QStringList dirsInDir2 = dir2.entryList( emptyFileFilter, QDir::AllDirs | mEntryListFilter );

	foreach( QString subDir, dirsInDir1 )
	{
		if ( isIgnored( subDir ) )
			continue;

		if( dirsInDir2.contains( subDir ) )
		{
			// Compare the common sub directory
			comparePrivate( (directory1 + "\\" + subDir), (directory2 + "\\" + subDir) );
		
			// remove subDir from dirsInDir2 so we know its been found.
			dirsInDir2.removeAt( dirsInDir2.indexOf( subDir ) );
		}
		else
		{
			// directory has been removed.
			directoryRemoved( directory1 + "\\" + subDir );
		}
	}

	//
	// Now check if dirsInDir2 still contains any directories. These are new.
	//
	foreach( QString subDir, dirsInDir2 )
	{
		if ( isIgnored( subDir ) )
			continue;

		directoryAdded( directory2 + "\\" + subDir );
	}

}




void DirectoryDiff::compareFilesIn( const QString& directory1, const QString& directory2 )
{
	QDir dir1( directory1 );
	QDir dir2( directory2 );

	if( !dir1.exists() ) throw QString( directory1 + " not found." );
	if( !dir2.exists() ) throw QString( directory2 + " not found." );

	QStringList fileFilter;
	fileFilter.append("*");

	QStringList filesInDir1 = dir1.entryList( fileFilter, QDir::Files | mEntryListFilter );
	QStringList filesInDir2 = dir2.entryList( fileFilter, QDir::Files | mEntryListFilter );

	foreach( QString file, filesInDir1 )
	{
		if ( isIgnored( file ) )
			continue;

		if( filesInDir2.contains( file ) )
		{
			// Compare the common file.
			compareFile( directory1 + "\\" + file, directory2 + "\\" + file );
		
			// remove the file from filesInDir2 so we know its been found.
			filesInDir2.removeAt( filesInDir2.indexOf( file ) );
		}
		else
		{
			// file has disappeared.
			addToList( "File Removed", mFilesRemoved, directory1 + "\\" + file );
		}
	}

	//
	// Now check if filesInDir2 still contains any files. These are new.
	//
	foreach( QString file, filesInDir2 )
	{
		if ( isIgnored( file ) )
			continue;

		addToList( "File Added", mFilesAdded, directory2 + "\\" + file );
	}

}


void DirectoryDiff::compareFile( const QString& filename1, const QString& filename2 )
{
	QFile file1( filename1 );
	QFile file2( filename2 );

	Q_ASSERT( file1.exists() );
	Q_ASSERT( file2.exists() );

	if( ! file1.open(QIODevice::ReadOnly) ) throw QString( "Unable to open " + filename1 + " for reading." );
	if( ! file2.open(QIODevice::ReadOnly) ) throw QString( "Unable to open " + filename2 + " for reading." );

	//Check the sizes of the two files...

	const qint64 file1Size = file1.size();
	const qint64 file2Size = file2.size();

	bool filesDifferent = ( file1Size != file2Size );

	if ( !filesDifferent )
	{
		const int STANDARD_CHUNK_SIZE = 1000000;
		const int MAX_CHUNK_SIZE = qMin( file1Size, qint64( STANDARD_CHUNK_SIZE ) );
		qint64 bytesLeft = file1Size;
		
		QByteArray file1Contents(MAX_CHUNK_SIZE, 0);
		QByteArray file2Contents(MAX_CHUNK_SIZE, 0);

		int currentChunkSize;
		while ( !filesDifferent && (bytesLeft > 0) )
		{

			bytesLeft -= MAX_CHUNK_SIZE;
			if (bytesLeft >= 0)
			{
				currentChunkSize = MAX_CHUNK_SIZE;
			}
			else
			{
				currentChunkSize = bytesLeft + MAX_CHUNK_SIZE;
				file1Contents.resize( currentChunkSize );
				file2Contents.resize( currentChunkSize );

			}
			file1.read( file1Contents.data(), currentChunkSize );
			file2.read( file2Contents.data(), currentChunkSize );

			filesDifferent = ( file1Contents != file2Contents );

		}
	}
	if( filesDifferent )
	{
		addToList( "File Modified", mFilesModified, filename1 );
	}
}


void DirectoryDiff::directoryRemoved( const QString& directory1 )
{
	QDir dir1(directory1);
	Q_ASSERT( dir1.exists() );

	addToList( "Directory Removed", mDirectoriesRemoved, directory1 );

	QStringList fileFilter;
	fileFilter.append("*");

	QStringList filesInDir1 = dir1.entryList( fileFilter, QDir::Files | mEntryListFilter );
	static const QRegExp START_OF_FILENAME( "^" );
	addToList( "File Removed", mFilesRemoved, filesInDir1.replaceInStrings( START_OF_FILENAME, directory1 + "\\" ) );

	QStringList dirsInDir1 = dir1.entryList( QDir::AllDirs | mEntryListFilter );
	foreach( QString subdir, dirsInDir1 )
	{
		directoryRemoved( directory1 + "\\" + subdir );
	}
}


void DirectoryDiff::directoryAdded( const QString& directory2 )
{
	QDir dir2(directory2);
	Q_ASSERT( dir2.exists() );

	addToList( "Directory Added", mDirectoriesAdded, directory2 );

	QStringList fileFilter;
	fileFilter.append("*");

	QStringList filesInDir2 = dir2.entryList( fileFilter, QDir::Files | mEntryListFilter );
	static const QRegExp START_OF_FILENAME( "^" );
	addToList( "File Added", mFilesAdded, filesInDir2.replaceInStrings( START_OF_FILENAME, directory2 + "\\" ) );

	QStringList dirsInDir2 = dir2.entryList( QDir::AllDirs | mEntryListFilter );
	foreach( QString subdir, dirsInDir2 )
	{
		directoryAdded( directory2 + "\\" + subdir );
	}
}


const QStringList& DirectoryDiff::filesAdded() const
{
	return mFilesAdded;
}

const QStringList& DirectoryDiff::filesRemoved() const
{
	return mFilesRemoved;
}

const QStringList& DirectoryDiff::filesModified() const
{
	return mFilesModified;
}

const QStringList& DirectoryDiff::directoriesAdded() const
{
	return mDirectoriesAdded;
}

const QStringList& DirectoryDiff::directoriesRemoved() const
{
	return mDirectoriesRemoved;
}

void DirectoryDiff::addToList( const QString& label, QStringList& list, const QStringList& items )
{
	foreach( QString item, items )
	{
		addToList( label, list, item );
	}
}

void DirectoryDiff::addToList( const QString& label, QStringList& list, const QString& item )
{
	/// Remove the root path from the file name/directory name.
	QString itemLocal = item;
	if( item.startsWith( mDirectory1 + "\\" ) )
	{
		itemLocal = itemLocal.remove( mDirectory1 + "\\" );
	}
	else if( item.startsWith( mDirectory2 + "\\" ) )
	{
		itemLocal = itemLocal.remove( mDirectory2 + "\\" );
	}

	log( label, itemLocal );
	list.append( itemLocal );
}


void DirectoryDiff::log( const QString& label, const QStringList& list )
{
	foreach( QString item, list )
	{
		log( label, item );
	}
}

void DirectoryDiff::log( const QString& label, const QString& item )
{
	QString prefix = label + ":";
	qDebug( "%-25s %s", prefix, item );
}


void DirectoryDiff::includeSystem( bool enable )
{
	if( enable )
	{
		mEntryListFilter |= QDir::System;
	}
	else
	{
		mEntryListFilter &= (~QDir::System);
	}
}


void DirectoryDiff::includeHidden( bool enable )
{
	if( enable )
	{
		mEntryListFilter |= QDir::Hidden;
	}
	else
	{
		mEntryListFilter &= (~QDir::Hidden);
	}
}

bool DirectoryDiff::isIgnored( QString path )
{
	path = path.remove( mBasePathRemover );
	foreach( QRegExp mask, mIgnoreList )
	{
		if ( mask.exactMatch( path ) )
			return true;
	}

	return false;
}
