/*

nsisPatchGen
Copyright (C) 2006  Vibration Technology Ltd. ( http://www.vibtech.co.uk )

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*/

/**
  * \file
**/

#include "StartExternals.h"
#include <QTextStream>
#include <QStringList>
#include <QFileInfo>
#include "StopExternals.h"

#include "NsisGenerator.h"

const QString INDENT( "  " );
const QString ENDL( "\r\n" ); // mickeysoft stylee!!

NsisGenerator::NsisGenerator( const QString& outputFileName )
: mOutFile( outputFileName )
{

}

NsisGenerator::~NsisGenerator( )
{
	close();
}

void NsisGenerator::open( ) throw( QString )
{
	if( ! mOutFile.open( QIODevice::WriteOnly ) )
	{
		throw QString( "Failed to open the NSIS output file: \"" + mOutFile.fileName() + "\"" );
	}

	mOut.setDevice( &mOutFile );

	writeHeader( );
}

void NsisGenerator::writeHeader( )
{
	//Include our own version of the VPatchFile macro
	static const QString NSISPG_VPATCH_MACRO = ""
	"!macro NSISPGPatchFile PATCHDATA SOURCEFILE TEMPFILE\r\n"
	"  InitPluginsDir\r\n"
	"  File \"/oname=$PLUGINSDIR\\${TEMPFILE}.patchFile\" \"${PATCHDATA}\"\r\n"
	"  vpatch::vpatchfile \"$PLUGINSDIR\\${TEMPFILE}.patchFile\" \"${SOURCEFILE}\" \"${TEMPFILE}\"\r\n"
	"  Pop $1\r\n"
	"  DetailPrint $1\r\n"
	"  StrCpy $1 $1 2\r\n"
	"  StrCmp $1 \"OK\" +2\r\n"
	"  SetErrors\r\n"
	"  IfFileExists \"${TEMPFILE}\" +1 +3\r\n"
	"  Delete \"${SOURCEFILE}\"\r\n"
	"  Rename /REBOOTOK \"${TEMPFILE}\" \"${SOURCEFILE}\"\r\n"
	"  Delete \"$PLUGINSDIR\\${TEMPFILE}.patchFile\"\r\n"
	"!macroend\r\n";

	mOut << NSISPG_VPATCH_MACRO;

	mOut << ENDL;
}

void NsisGenerator::close( )
{
	mOutFile.close();
}

void NsisGenerator::filesAdded( const QStringList& files )
{
	mOut << "Function patchFilesAdded" << ENDL;

	foreach( QString filename, files )
	{
		QString file;
		QString directory;
		splitFileAndPath( filename, directory, file );

		mOut << INDENT << "SetOutPath \"${PATCH_INSTALL_ROOT}";
		if( ! directory.isEmpty() )
		{
			mOut << "\\" << directory;
		}
		mOut << "\"" << ENDL;

		mOut << INDENT << "File \"${PATCH_SOURCE_ROOT}\\" << filename << "\"" << ENDL;
	}

	mOut << "FunctionEnd" << ENDL << ENDL;
}

void NsisGenerator::filesRemoved( const QStringList& files )
{
	mOut << "Function patchFilesRemoved" << ENDL;

	foreach( QString filename, files )
	{
		mOut << INDENT << "Delete \"${PATCH_INSTALL_ROOT}\\" << filename << "\"" << ENDL;
	}

	mOut << "FunctionEnd" << ENDL << ENDL;
}

void NsisGenerator::filesModified( const QStringList& files )
{
	mOut << "Function patchFilesModified" << ENDL;
	mOut << INDENT << "InitPluginsDir" << ENDL << ENDL;

	foreach( QString fullFileName, files )
	{
		QString patchFile  = "${PATCH_FILES_ROOT}\\" + fullFileName + ".pat";
		QString destFile   = "${PATCH_INSTALL_ROOT}\\" + fullFileName;
		QString tempFileName = QFileInfo( fullFileName ).fileName() + ".tmp";
		

		mOut << INDENT << ";Patching \"" << fullFileName << "\" using patch file \"" << fullFileName + ".pat" << "\"" << ENDL;
		mOut << INDENT << "DetailPrint \"Patching " << destFile << "\"" << ENDL;

		
		mOut << INDENT << "!insertmacro NSISPGPatchFile " 
			 << "\""  << patchFile << "\" "
			 << "\""  << destFile << "\" "
			 << "\""  << tempFileName << "\"" 
			 << ENDL;

		mOut << ENDL;
	}

	mOut << "FunctionEnd" << ENDL << ENDL;
}

void NsisGenerator::directoriesAdded( const QStringList& dirs )
{
	mOut << "Function patchDirectoriesAdded" << ENDL;

	foreach( QString directoryName, dirs )
	{
		mOut << INDENT << "CreateDirectory \"${PATCH_INSTALL_ROOT}\\" << directoryName << "\"" << ENDL;
	}

	mOut << "FunctionEnd" << ENDL << ENDL;
}


void NsisGenerator::directoriesRemoved( const QStringList& dirs )
{
	mOut << "Function patchDirectoriesRemoved" << ENDL;

	foreach( QString directoryName, dirs )
	{
		mOut << INDENT << "RMDir /r \"${PATCH_INSTALL_ROOT}\\" << directoryName << "\"" << ENDL;
	}

	mOut << "FunctionEnd" << ENDL << ENDL;
}

void NsisGenerator::splitFileAndPath( const QString& fullPath, QString& dirOut, QString& fileOut )
{
	int lastSeparator = fullPath.lastIndexOf('\\');

	if( -1 != lastSeparator )
	{
		dirOut = fullPath.left(lastSeparator);
		fileOut = fullPath.right(fullPath.size() - lastSeparator - 1);
	}
	else
	{
		dirOut = "";
		fileOut = fullPath;
	}
}



