/*

nsisPatchGen
Copyright (C) 2006  Vibration Technology Ltd. ( http://www.vibtech.co.uk )

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "StartExternals.h"
#include <QProcess>
#include "StopExternals.h"

#include "DirectoryDiffTest.h"

#include "DirectoryDiff.h"


// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( DirectoryDiffTest );

const QString DIR1("testData\\DirectoryDiffTest\\dir1");
const QString DIR2("testData\\DirectoryDiffTest\\dir2");
const QString DIR3("testData\\DirectoryDiffTest\\dir10");

void DirectoryDiffTest::setUp()
{
	setFileAttributes();
}

void DirectoryDiffTest::tearDown()
{
}

void DirectoryDiffTest::testConstructor()
{
	DirectoryDiff directoryDiff;
	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() == 0 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() == 0 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() == 0 );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().size() == 0 );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().size() == 0 );

}


void DirectoryDiffTest::testCompareFilesIn()
{
	DirectoryDiff directoryDiff;
	try
	{
		directoryDiff.compareFilesIn( DIR1, DIR2 );
	}
	catch( QString errorText )
	{
		qDebug() << "Caught unexpected exception from DirectoryDiff::compareFilesIn: " << endl << errorText << endl;
		CPPUNIT_ASSERT( false );
	}

	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() == 1 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() == 2 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() == 2 );

}



void DirectoryDiffTest::testCompare()
{
	DirectoryDiff directoryDiff;
	try
	{
		directoryDiff.compare( DIR1, DIR2 );
	}
	catch( QString errorText )
	{
		qDebug() << "Caught unexpected exception from DirectoryDiff::compare: " << endl << errorText << endl;
		CPPUNIT_ASSERT( false );
	}


	listFiles( "Files added", directoryDiff.filesAdded() );
	listFiles( "Files removed", directoryDiff.filesRemoved() );
	listFiles( "Files modified", directoryDiff.filesModified() );

	listFiles( "Directories added", directoryDiff.directoriesAdded() );
	listFiles( "Directories removed", directoryDiff.directoriesRemoved() );


	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() == 4 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() == 5 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() == 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().size() == 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().size() == 4 );

	verifyListOfNonHiddenNonSystemFiles( directoryDiff );
}

void DirectoryDiffTest::testCompareSimilar()
{
	DirectoryDiff directoryDiff;
	try
	{
		directoryDiff.compare( DIR1, DIR3 );
	}
	catch( QString errorText )
	{
		qDebug() << "Caught unexpected exception from DirectoryDiff::compare: " << endl << errorText << endl;
		CPPUNIT_ASSERT( false );
	}


	listFiles( "Files added", directoryDiff.filesAdded() );
	listFiles( "Files removed", directoryDiff.filesRemoved() );
	listFiles( "Files modified", directoryDiff.filesModified() );

	listFiles( "Directories added", directoryDiff.directoriesAdded() );
	listFiles( "Directories removed", directoryDiff.directoriesRemoved() );


	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() == 4 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() == 5 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() == 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().size() == 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().size() == 4 );

	verifyListOfNonHiddenNonSystemFiles( directoryDiff );
}

void DirectoryDiffTest::listFiles( const QString & type, const QStringList& files )
{
	qDebug() << type << "(" << files.size() << " files):";
	foreach( QString file, files )
	{
		qDebug() << "  " << file;
	}
}

void DirectoryDiffTest::testCompareFailures()
{
	DirectoryDiff diff;
	const QString INVALID_DIR1( "this\\is\\an\\invalid\\directory" );
	const QString INVALID_DIR2( "and\\so\\is\\this" );
	const QString VALID_DIR("testData\\DirectoryDiffTest\\dir1");

	bool exceptionCaught = false;

	try
	{
		diff.compare( INVALID_DIR1, INVALID_DIR2 );
		exceptionCaught = false;
	}
	catch( QString error )
	{
		CPPUNIT_ASSERT( error.contains( INVALID_DIR1 ) );
		exceptionCaught = true;
	}
	CPPUNIT_ASSERT( exceptionCaught );

	
	try
	{
		diff.compare( VALID_DIR, INVALID_DIR2 );
		exceptionCaught = false;
	}
	catch( QString error )
	{
		CPPUNIT_ASSERT( error.contains( INVALID_DIR2 ) );
		exceptionCaught = true;
	}
	CPPUNIT_ASSERT( exceptionCaught );


	try
	{
		diff.compare( INVALID_DIR1, VALID_DIR );
		exceptionCaught = false;
	}
	catch( QString error )
	{
		CPPUNIT_ASSERT( error.contains( INVALID_DIR1 ) );
		exceptionCaught = true;
	}
	CPPUNIT_ASSERT( exceptionCaught );

}



void DirectoryDiffTest::testCompareIncludeHiddenFiles()
{
	DirectoryDiff directoryDiff;
	directoryDiff.includeHidden( true );

	qDebug() << "*************** INCLUDE HIDDEN FILES ****************";

	try
	{
		directoryDiff.compare( DIR1, DIR2 );
	}
	catch( QString errorText )
	{
		qDebug() << "Caught unexpected exception from DirectoryDiff::compare: " << endl << errorText << endl;
		CPPUNIT_ASSERT( false );
	}


	listFiles( "Files added", directoryDiff.filesAdded() );
	listFiles( "Files removed", directoryDiff.filesRemoved() );
	listFiles( "Files modified", directoryDiff.filesModified() );

	listFiles( "Directories added", directoryDiff.directoriesAdded() );
	listFiles( "Directories removed", directoryDiff.directoriesRemoved() );


	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() >= 4 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() > 5 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() >= 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().size() >= 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().size() >= 4 );

	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "removedDir\\.removed"  ) );

	verifyListOfNonHiddenNonSystemFiles( directoryDiff );
}



void DirectoryDiffTest::testCompareIncludeSystemFiles()
{
	DirectoryDiff directoryDiff;
	directoryDiff.includeSystem( true );
	directoryDiff.includeHidden( true );

	qDebug() << "*************** INCLUDE HIDDEN SYSTEM FILES ****************";

	try
	{
		
		directoryDiff.compare( DIR1, DIR2 );
	}
	catch( QString errorText )
	{
		qDebug() << "Caught unexpected exception from DirectoryDiff::compare: " << endl << errorText << endl;
		CPPUNIT_ASSERT( false );
	}


	listFiles( "Files added", directoryDiff.filesAdded() );
	listFiles( "Files removed", directoryDiff.filesRemoved() );
	listFiles( "Files modified", directoryDiff.filesModified() );

	listFiles( "Directories added", directoryDiff.directoriesAdded() );
	listFiles( "Directories removed", directoryDiff.directoriesRemoved() );

	CPPUNIT_ASSERT( directoryDiff.filesAdded().size() >= 4 );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().size() > 5 );
	CPPUNIT_ASSERT( directoryDiff.filesModified().size() >= 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().size() >= 3 );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().size() >= 4 );

	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "Thumbs.db"  ) );

	verifyListOfNonHiddenNonSystemFiles( directoryDiff );
}

void DirectoryDiffTest::verifyListOfNonHiddenNonSystemFiles( const DirectoryDiff& directoryDiff )
{
	CPPUNIT_ASSERT( directoryDiff.filesAdded().contains( "changedDir\\newFiles\\newFile.txt" ) );
	CPPUNIT_ASSERT( directoryDiff.filesAdded().contains( "newDir\\newDirWithFiles\\newFile.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesAdded().contains( "newDir\\newDirWithFiles\\newFile2.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesAdded().contains( "newFile.txt"  ) );

	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "changedDir\\removedFiles\\removedFile.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "removedDir\\removedWithFiles\\removedFile.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "removedDir\\removedWithFiles\\removedFile2.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesRemoved().contains( "removedFile.txt"  ) );

	CPPUNIT_ASSERT( directoryDiff.filesModified().contains( "changedDir\\changedFiles\\changedFile.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesModified().contains( "changedFile.txt"  ) );
	CPPUNIT_ASSERT( directoryDiff.filesModified().contains( "changedFile.bin"  ) );

	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().contains( "newDir"  ) );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().contains( "newDir\\newDirNoFiles" ) );
	CPPUNIT_ASSERT( directoryDiff.directoriesAdded().contains( "newDir\\newDirWithFiles" ) );

	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().contains( "removedDir" ) );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().contains( "removedDir\\removedNoFiles" ) );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().contains( "removedDir\\removedWithFiles" ) );
	CPPUNIT_ASSERT( directoryDiff.directoriesRemoved().contains( "removedDirEmpty" ) );
}

void DirectoryDiffTest::setFileAttributes()
{
	QProcess::execute("attrib +S " + DIR1 + "\\Thumbs.db");
	QProcess::execute("attrib +H " + DIR1 + "\\removedDir\\.removed");
}
